//
// Created by vout on 11/21/18.
//

#ifndef MEGAHIT_ALL_ALGO_H
#define MEGAHIT_ALL_ALGO_H

#include "bubble_remover.h"
#include "low_depth_remover.h"
#include "sdbg_pruning.h"
#include "tip_remover.h"
#include "weak_link_remover.h"

#endif  // MEGAHIT_ALL_ALGO_H
